from __future__ import absolute_import, division, print_function
import numpy as np
import rospy
from scipy.spatial import cKDTree
from .utils import affine_transform


def icp(x, y, y_index=None, max_iters=50, inlier_ratio=1.0, inlier_dist_mult=1.0, rel_tol=1e-3, T=None):
    """Iterative closest point algorithm, minimizing sum of squares of point
    to point distances.

    :param x: Points to align, D-by-M array.
    :param y: Reference points to align to, D-by-N array, such that
            [y[:, j]; 1] is approximately T*[x[:, i]; 1] for corresponding
            pair (i, j). The correspondences are established from nearest
            neighbors (i.e., "closest points").
    :param y_index: Index for NN search for y (cKDTree)
            which can be queried with row 1-by-D vectors.
    :param max_iters: Maximum number of iterations.
    :param inlier_ratio: Ratio of inlier correspondences with lowest
            nearest-neighbor distances for which we optimize the criterion
            in given iteration. This should correspond to the minimum relative
            overlap between point clouds. The inliers set may change each
            iteration.
    :param inlier_dist_mult: Multiplier of the maximum inlier distance found
            using inlier ratio above, enlarging or reducing the inlier set for
            optimization.
    :param rel_tol: Minimum relative improvement in each iteration to
           continue.
    :param T: Initial transform estimate from SE(D), defaults to identity.
    :return: Optimized transform from SE(D),
            mean inlier error (L2 distance) from the last iteration, and
            boolean inlier mask from the last iteration, x[:, inl] are the
            inlier points.
    """
    assert x.shape[0] == y.shape[0]
    if y_index is None:
        y_index = cKDTree(y.T)
    if T is None:
        T = np.eye(x.shape[0] + 1)
    # Boolean inlier mask from current iteration.
    inl = None
    # Mean inlier distance from current iteration.
    inl_err = None
    # Mean inlier distance from previous iteration (to assess improvement).
    prev_inl_err = None
    # TODO: Implement!

    return T, inl_err, inl
