from __future__ import absolute_import, division, print_function
from geometry_msgs.msg import Point, Transform, TransformStamped
from nav_msgs.msg import OccupancyGrid
import numpy as np
from ros_numpy import msgify, numpify
from voxel_map import VoxelMap
from sensor_msgs.msg import PointCloud2


def slots(msg):
    """Return message attributes (slots) as list."""
    return [getattr(msg, var) for var in msg.__slots__]


def array(msg):
    """Return message attributes (slots) as array."""
    return np.array(slots(msg))


def col(arr):
    """Convert array to column vector."""
    return arr.reshape((arr.size, 1))


def logistic(x):
    """Standard logistic function, inverse of logit function."""
    return 1. / (1. + np.exp(-x))


def logit(p):
    """Logit function or the log-odds, inverse of logistic function.
    The logarithm of the odds p / (1 - p) where p is probability."""
    return np.log(p / (1. - p))


def affine_transform(T, x):
    """Affine transform T * [x; 1]. The last row in T is neglected."""
    assert T.ndim == 2
    assert x.ndim == 2
    d = x.shape[0]
    assert d <= T.shape[0] <= d + 1
    assert T.shape[1] == d + 1
    R = T[:d, :d]
    t = T[:d, d:]
    return np.matmul(R, x) + t


def inverse_affine(T):
    assert T.ndim == 2
    d = T.shape[1] - 1
    assert d <= T.shape[0] <= d + 1
    R = T[:d, :d]
    t = T[:d, d:]
    T_inv = T.copy()
    T_inv[:d, :d] = R.T
    T_inv[:d, d:] = -np.matmul(R.T, t)
    return T_inv


def rotation_angle(R):
    assert R.ndim == 2
    assert R.shape[0] == R.shape[1]
    alpha = np.arccos((np.trace(R) - (R.shape[0] - 2)) / 2.)
    return alpha


def msg_to_cloud(msg, fields=('x', 'y', 'z')):
    """Convert PointCloud2 message to F-by-N array using given F fields."""
    x = numpify(msg).ravel()
    x = np.stack([x[f] for f in fields])
    return x


def cloud_to_msg(x, fields=('x', 'y', 'z')):
    """Convert cloud F-by-N array to PointCloud2 message."""
    dtype = list(zip(fields, len(fields) * [np.float32]))
    cloud = np.zeros(x.shape[1], dtype=dtype)
    for i, f in enumerate(fields):
        cloud[f] = x[i, :]
    msg = msgify(PointCloud2, cloud)
    return msg


def msg_to_transform(msg):
    """Convert TransformStamped message to 4-by-4 transform matrix."""
    return numpify(msg.transform)


def transform_to_msg(T):
    """Convert 4-by-4 transform matrix to TransformStamped message."""
    msg = TransformStamped()
    msg.transform = msgify(Transform, T)
    return msg


def filter_grid(x, grid_res, keep='random'):
    """Select random point within each cell. Order is not preserved."""
    # Convert points in cols to rows, and shuffle the rows.
    if keep == 'first':
        # Make the first item last.
        x = x[:, ::-1].T
    elif keep == 'random':
        # Make the last item random.
        x = x.T.copy()
        np.random.shuffle(x)
    elif keep == 'last':
        # Keep the last item last.
        pass
    # Get integer cell indices, as tuples.
    idx = np.floor(x / grid_res)
    idx = [tuple(i) for i in idx]
    # Dict keeps the last value for each key, which set above by keep param.
    x = dict(zip(idx, x))
    # Concatenate and convert rows to cols.
    x = np.stack(x.values()).T
    return x


def zeropad(x, shape):
    assert np.all(np.less_equal(x.shape, shape))
    pad = [(0, n - m) for m, n in zip(x.shape, shape)]
    x = np.pad(x, pad, 'constant')
    return x


class PointMap(object):
    def __init__(self, grid_res=None, max_size=None):
        self.cloud = None
        self.grid_res = grid_res
        self.max_size = max_size

    def size(self):
        if self.cloud is None:
            return 0
        return self.cloud.shape[1]

    def empty(self):
        return self.size() == 0

    def update(self, x):
        if self.empty():
            self.cloud = x
        else:
            self.cloud = np.concatenate((self.cloud, x), axis=1)
        if self.grid_res is not None:
            self.cloud = filter_grid(self.cloud, self.grid_res, keep='first')
        if self.max_size is not None and self.size() > self.max_size:
            self.cloud = self.cloud[:, -self.max_size:]


class OccupancyMap(object):
    def __init__(self, frame_id, resolution=0.1, min=-10.0, max=10.0, occupied=5.0):
        self.voxel_map = VoxelMap(resolution, -1.0, 1.0, 3.0)
        self.min = min
        self.max = max
        self.msg = OccupancyGrid()
        self.msg.header.frame_id = frame_id
        self.msg.info.resolution = resolution
        # Initially, set the grid origin to identity.
        self.msg.info.origin.orientation.w = 1.0

    def map_to_grid(self, x):
        """Transform points from map coordinates to grid."""
        # TODO: Handle orientation too.
        x = x - col(array(self.msg.info.origin.position))
        return x

    def grid_to_map(self, x):
        """Transform points from grid coordinates to map."""
        # TODO: Handle orientation too.
        x = x + col(array(self.msg.info.origin.position))
        return x

    def fit_grid(self):
        """Accommodate the grid to contain all points."""
        # Update grid origin so that all coordinates are non-negative.
        x, _, v = self.voxel_map.get_voxels()
        if x.size == 0:
            return
        x = x[:2]  # Only x,y used in 2D grid.
        x_min = x.min(axis=1) - self.voxel_map.voxel_size / 2.
        x_max = x.max(axis=1) + self.voxel_map.voxel_size / 2.
        nx = np.round((x_max - x_min) / self.msg.info.resolution).astype(np.int)
        self.msg.info.origin.position = Point(x_min[0], x_min[1], 0.0)
        self.msg.info.width, self.msg.info.height = nx

    def grid_voxels(self):
        """Return voxel coordinates corresponding to the current grid."""
        i, j = np.meshgrid(np.arange(self.msg.info.width),
                           np.arange(self.msg.info.height),
                           indexing='xy')
        x = np.stack((i.ravel(), j.ravel(), np.zeros_like(i).ravel()))
        x = (x + 0.5) * self.msg.info.resolution
        return x

    def to_msg(self):
        """Return as occupancy grid message. Update grid parameters as needed."""
        self.fit_grid()
        x = self.grid_voxels()
        x = self.grid_to_map(x)
        x[2, :] = self.voxel_map.voxel_size / 2.0
        l = np.zeros((x.shape[1],))
        v = self.voxel_map.get_voxels(x, l)
        v = 100. * logistic(v)
        v[np.isnan(v)] = -1.
        self.msg.data = v.astype(int).tolist()
        return self.msg
    
    def voxel_map_points(self, x):
        x = x.copy()
        x[2, :] = self.voxel_map.voxel_size / 2.0
        return x

    def update(self, x, y, stamp):
        """Update internal occupancy map."""
        x = self.voxel_map_points(x)
        y = self.voxel_map_points(y)
        if x.shape[1] == 1:
            x = np.broadcast_to(x, y.shape)
        elif y.shape[1] == 1:
            y = np.broadcast_to(y, x.shape)
        self.voxel_map.update_lines(x, y)
        self.clip_values()
        self.msg.header.stamp = stamp
        self.msg.info.map_load_time = stamp
    
    def occupied(self, x):
        x = self.voxel_map_points(x)
        l = np.zeros((x.shape[1],))
        v = self.voxel_map.get_voxels(x, l)
        occupied = v > self.voxel_map.occupied_threshold
        return occupied
    
    def clip_values(self):
        x, l, v = self.voxel_map.get_voxels()
        v = np.clip(v, self.min, self.max)
        self.voxel_map.set_voxels(x, l, v)
